#version 430 core

uniform mat4 vmMatrix;
uniform mat3 vmNormalMatrix;

uniform float uNormalMappingStrength = 1.0;

layout(binding = 0) uniform sampler2D uDiffuseTexSampler;
layout(binding = 1) uniform sampler2D uNormalTexSampler;

layout(binding = 10) uniform samplerCube shadowCubeMap;

struct MaterialData {
	int type;

	vec3 ambientReflectivity;
	vec3 diffuseReflectivity;
	vec3 specularReflectivity;
	float shininess;

	float alpha;
	vec3 F0;
	float metalness;
};

struct AmbientLightData
{
	vec4 color;
};

struct DirectionalLightData
{
	vec4 color;
	vec4 direction;
};

struct PointLightData
{	
	vec4 color;
	vec3 position;
};

uniform vec3 burstPosition = vec3(-0.850889709825530, 0.168937838383838, 0.566278877869604); //vec3(-0.940057734439834, 0.158330933609959, 0.568162412863070);
uniform float burstRadius = 0.9;
uniform float burstSave = 0.165;

uniform MaterialData uMaterial;
uniform AmbientLightData uAmbientLight;
uniform DirectionalLightData uDirectionalLight;
uniform PointLightData uPointLight;

in VertexData
{
	vec3 positionOS;
	vec3 positionVS; 
	vec2 uvCoords;   
	vec3 normalOS;   
	vec3 tangentOS;  
	vec3 bitangentOS;
} fs_in;

out vec4 oFragColor;



// Shadow Stuff-----------------------------------------------------------------

int numSamples = 20;

vec3 offsetDirections[20] = vec3[]
(
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
);

vec3 ReOrthogonalize(vec3 first, vec3 second)
{
	return normalize(first - dot(first, second) * second);
}

vec3 CalculateNormal()
{
	vec3 normalOS = normalize(fs_in.normalOS);
	vec3 tangentOS = ReOrthogonalize(fs_in.tangentOS, normalOS);
	vec3 bitangentOS = ReOrthogonalize(fs_in.bitangentOS, normalOS);

	mat3 matrixTStoOS = mat3(tangentOS, bitangentOS, normalOS); // TBN matrix

	vec3 normalTEX = normalize(texture(uNormalTexSampler, fs_in.uvCoords).rgb * 2.0 - 1.0);
	normalTEX.xy *= uNormalMappingStrength;
	vec3 normalVS = vmNormalMatrix * matrixTStoOS * normalTEX;

	return normalize(normalVS);
}

float CalculateShadow(vec3 vertexPosition) {
	
	float shadow = 0.0;
	float farPlane = 25.0;
	vec3 normal = CalculateNormal();

	vec3 directionFromLight = vertexPosition - uPointLight.position; 

	float radius = (1.0 + (length(directionFromLight) / farPlane)) / 300.0; 

	float bias = max(0.14 * (1.0 - dot(normal, directionFromLight)), 0.014);

	for (int i = 0; i < numSamples; i++){
		float nearestDepth = texture(shadowCubeMap, normalize(directionFromLight) + offsetDirections[i] * radius).r;	
		nearestDepth = nearestDepth * farPlane;
		
		if ((length(directionFromLight) - 0.014) > nearestDepth) {
			shadow = shadow + 1.0;
		}
	}

	return (shadow / numSamples);
}

vec3 CalculateBlinnPhongShading(vec3 light, vec3 eye, vec3 normal, vec3 diffuseColor)
{
	float n_dot_l = max(0.0, dot(normal, light));
	vec3 H = normalize(light + eye);
	float n_dot_h = max(0.0, dot(normal, H));
	float s = pow(n_dot_h, uMaterial.shininess);
	
	vec3 diffuse = uMaterial.diffuseReflectivity * diffuseColor * n_dot_l;
	vec3 specular = uMaterial.specularReflectivity * s;

	return diffuse + specular;
}

// PBS


// Fresnel's Law
vec3 F(vec3 l, vec3 h)
{
	float l_dot_h = max(0.0, dot(l, h));
	return uMaterial.F0 + (1 - uMaterial.F0) * pow(1 - l_dot_h, 5);
}

// Fraction of Surface oriented along half-way vector
float D(vec3 n, vec3 m)
{
	float n_dot_m = max(0.0, dot(n, m));
	return (uMaterial.alpha * uMaterial.alpha) / (3.14159 * pow(n_dot_m * n_dot_m * (uMaterial.alpha * uMaterial.alpha - 1) + 1, 2));
}

float Lambda(float r)
{
	float root = 1 + uMaterial.alpha * uMaterial.alpha * (1 - r * r) / (r * r);
	return (-1 + sqrt(root)) / 2;
}

// Fraction of unoccluded rays (Correlated Smith)
float G(vec3 l, vec3 v, vec3 n)
{
	float n_dot_l = max(0.0, dot(n, l));
	float v_dot_n = max(0.0, dot(v, n));
	return 1 / (1 + Lambda(n_dot_l) + Lambda(v_dot_n));
}

// Calculate Physicall Based Shading with Microfacet model BRDF
vec3 CalculatePhysicallyBasedShading(vec3 light, vec3 eye, vec3 normal)
{
	float n_dot_l = max(0.0, dot(normal, light));
	float l_dot_v = max(0.0, dot(light, eye));
	float n_dot_v = max(0.0, dot(normal, eye));
	vec3 h = normalize(light + eye);
	if ((n_dot_l > 0) && (n_dot_v > 0)) {
		return F(light, h) * vec3(G(light, eye, normal) * D(normal, h) / (4 * n_dot_v * n_dot_l));
	} else {
		return vec3(0);
	}
}

vec3 CalculateDiffuseAndSpecular(vec3 diffuseColor, vec3 positionOS)
{
	vec3 result = vec3(0.0);

	vec3 pos = fs_in.positionVS;
	bool inside = (length(positionOS - burstPosition) < burstRadius) && (length(positionOS - burstPosition) > burstSave);

	vec3 eye = normalize(vec3(0.0) - pos); // VS -> eye 0,0,0
	vec3 light = -vec3(vmMatrix * uDirectionalLight.direction);
	vec3 normal = CalculateNormal();

	if (!inside) {
		result += uDirectionalLight.color.rgb * CalculateBlinnPhongShading(light, eye, normal, diffuseColor);
	} else {
		result += uDirectionalLight.color.rgb * CalculatePhysicallyBasedShading(light, eye, normal);
	}

	//normalize(vertexPosition - uPointLight.position)
	
	return result;
}

vec3 CalculatePointLighting(vec3 diffuseColor, vec3 positionOS)
{
	vec3 result = vec3(0.0);

	vec3 directionToLight = positionOS - uPointLight.position;
	vec3 directionToLightNormalized = normalize(directionToLight);

	float distanceSquared = dot(directionToLight, directionToLight);
	float distance = sqrt(distanceSquared);
	float distancePowThree = distance * distance * distance;
	
	float attenuation = 0.5 + 0.5 * distance + 0 * distanceSquared + 8 * distancePowThree;
	vec3 intensity = uPointLight.color.rgb / attenuation;

	vec3 pos = fs_in.positionVS;
	vec3 eye = normalize(vec3(0.0) - pos); // VS -> eye 0,0,0
	vec3 light = -vec3(vmMatrix * vec4(directionToLightNormalized, 0.0));
	vec3 normal = CalculateNormal();
	bool inside = (length(positionOS - burstPosition) < burstRadius) && (length(positionOS - burstPosition) > burstSave);	
	if (!inside) {
		result += intensity * CalculateBlinnPhongShading(light, eye, normal, diffuseColor);
	} else {
		result += intensity * CalculatePhysicallyBasedShading(light, eye, normal);
	}

	return result;
}

void main()
{
	float shadow = CalculateShadow(fs_in.positionOS);

	vec3 diffuseColor = texture(uDiffuseTexSampler, fs_in.uvCoords).rgb;

	vec3 ambient = uAmbientLight.color.rgb * uMaterial.ambientReflectivity * diffuseColor;
	vec3 diffuse_and_specular = CalculateDiffuseAndSpecular(diffuseColor, fs_in.positionOS);
	diffuse_and_specular += (1.0 - shadow) * CalculatePointLighting(diffuseColor, fs_in.positionOS);

	oFragColor = vec4(ambient + diffuse_and_specular, 1.0);
}